/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "grass/grass.pmh"

using namespace fe;

void SpineFitOp::initialize(void)
{
	catalog<String>("Facing Attribute")="facing";
	catalog<String>("Facing Attribute","hint")=
			"Name of vector attribute on input curves"
			" describing the twist orientation,"
			" preferably a normal presenting the widest exposure.";

	catalog<I32>("Samples")=5;
	catalog<I32>("Samples","min")=2;
	catalog<I32>("Samples","high")=10;
	catalog<I32>("Samples","max")=100;
	catalog<String>("Samples","hint")="Number of points along each curve.";

	catalog<I32>("Refinement")=0;
	catalog<I32>("Refinement","high")=4;
	catalog<I32>("Refinement","max")=8;
	catalog<String>("Refinement","hint")="Maximum level of subdivision."
			"  Refinement may be adaptive and non-uniform.";

	catalog<bool>("Fast Snap")=false;
	catalog<String>("Fast Snap","hint")=
			"Force sample vertices to colocate with existing surface vertices."
			"  This may be faster but probably much less evenly placed.";

	catalog<bool>("Debug Surface")=false;
	catalog<String>("Debug Surface","hint")=
			"Draws lines over the output showing the surface refinement."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	catalog<bool>("Debug Facing")=false;
	catalog<String>("Debug Facing","hint")=
			"Draws lines over the output showing the facing direction."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	catalog< sp<Component> >("Input Mesh");

	catalog< sp<Component> >("Output Curves");
	catalog<String>("Output Curves","IO")="output";

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void SpineFitOp::handle(Record& a_rSignal)
{
//	feLog("SpineFitOp::handle\n");

	const String facingName=catalog<String>("Facing Attribute");
	const I32 samples=catalog<I32>("Samples");
	const I32 refinement=catalog<I32>("Refinement");
	const BWORD fastSnap=catalog<bool>("Fast Snap");
	const BWORD debugSurface=catalog<bool>("Debug Surface");
	const BWORD debugFacing=catalog<bool>("Debug Facing");

	if(samples<2)
	{
		feLog("SpineFitOp::handle"
				" at least two samples required\n");
		catalog<String>("error")="samples must be >= 2";
		return;
	}

	sp<SurfaceAccessibleI> spOutputCurves;
	if(!accessOutput(spOutputCurves,a_rSignal)) return;

	sp<DrawI> spDrawOutput;
	if(!accessDraw(spDrawOutput,a_rSignal)) return;

	sp<DrawI> spDrawDebug;
	if(!accessGuide(spDrawDebug,a_rSignal)) return;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Mesh")) return;

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,spInputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,spOutputCurves,e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spAccessorFacing;
	access(spAccessorFacing,spOutputCurves,e_point,facingName,e_warning);

	const U32 count=spInputVertices->count();
	for(U32 index=0;index<count;index++)
	{
		sp<SurfaceI> spSubSurface;
		if(!access(spSubSurface,spInputAccessible,e_warning,index))
		{
			feLog("SpineFitOp::handle"
					" input mesh would not provide SurfaceI for %d\n", index);
			continue;
		}
		spSubSurface->setRefinement(refinement);

		SpatialVector* pCurve=new SpatialVector[samples];

		SpatialVector base;
		SpatialVector tip;
		set(base);
		set(tip);
		Real maxDist=0.0;

		const U32 subCount=spInputVertices->subCount(index);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const SpatialVector point=
					spInputVertices->spatialVector(index,subIndex);

			if(subIndex)
			{
				const Real distance=magnitude(point-base);
				if(maxDist<distance)
				{
					maxDist=distance;
					tip=point;
				}
			}
			else
			{
				base=point;
			}
		}

		pCurve[0]=base;
		pCurve[samples-1]=tip;

		for(I32 sample=1;sample<samples-1;sample++)
		{
			SpatialVector& rOutputPoint=pCurve[sample];

			rOutputPoint=base+(tip-base)*(sample/(samples-1.0));

			if(fastSnap)
			{
				//* brute force, on vertex only
				Real minDist=0.0;
				SpatialVector best=base;
				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const SpatialVector point=
							spInputVertices->spatialVector(index,subIndex);

					const Real distance=magnitude(point-rOutputPoint);
					if(!subIndex || minDist>distance)
					{
						minDist=distance;
						best=point;
					}
				}
				rOutputPoint=best;
			}
			else
			{
				try
				{
					sp<SurfaceI::ImpactI> spImpact=
							spSubSurface->nearestPoint(rOutputPoint);
					if(spImpact.isValid())
					{
						rOutputPoint=spImpact->intersection();
					}
				}
				catch(...)
				{
					feLog("SpineFitOp::handle nearestPoint failed"
							" (make sure input is triangulated)\n");
				}
			}
		}

		spDrawOutput->drawLines(pCurve,NULL,samples,DrawI::e_strip,false,NULL);

		spDrawDebug->pushDrawMode(m_spDebugGroup);

		if(spAccessorFacing.isValid() && samples>2)
		{
			const Color green(0.0,1.0,0.0,1.0);
			SpatialVector line[2];

			const I32 mid=samples>>1;
			const SpatialVector midpoint=pCurve[mid];

			const SpatialVector side=cross(midpoint-base,tip-midpoint);
			const SpatialVector facing=cross(side,tip-base);
			const Real mag=magnitude(facing);
			if(mag!=0.0)
			{
				const SpatialVector unitFacing=facing*(1.0/mag);

				//* TODO should be able to get a specific dir for each point
				for(I32 sample=0;sample<samples;sample++)
				{
					const U32 vertexIndex=
							spOutputVertices->integer(index,sample);
					spAccessorFacing->set(vertexIndex,unitFacing);

					if(debugFacing)
					{
						line[0]=pCurve[sample];
						line[1]=line[0]+facing;
						spDrawDebug->drawLines(line,NULL,2,DrawI::e_strip,
								false,&green);
					}
				}
			}
		}

		delete pCurve;

		if(debugSurface)
		{
			sp<DrawableI> spDrawable=spSubSurface;
			if(spDrawable.isValid())
			{
				spDrawable->draw(spDrawDebug,NULL);
			}

		}

		spDrawDebug->popDrawMode();
	}

//	feLog("SpineFitOp::handle done\n");
}
